# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula


from functools import partial
import importlib

from aeon.base.device import BaseDevice
from aeon.nxos.connector import NxosConnector as Connector


__all__ = ['Device']


class Device(BaseDevice):
    OS_NAME = 'nxos'

    def __init__(self, target, **kwargs):
        """
        :param target: hostname or ipaddr of target device
        :param kwargs:
            'user' : login user-name, defaults to "admin"
            'passwd': login password, defaults to "admin
        """
        BaseDevice.__init__(self, target, Connector, **kwargs)

    def close(self):
        # nothing to do for close at this time, yo!
        pass

    def gather_facts(self):
        exec_show = partial(self.api.exec_opcmd, resp_fmt='json')

        facts = self.facts

        got = exec_show('show hostname')
        facts['fqdn'] = got['hostname']
        facts['hostname'], _, facts['domain_name'] = facts['fqdn'].partition('.')

        attempts = 1
        while attempts < 5:
            try:
                got = exec_show('show hardware')
            except:
                attempts -= 1
            else:
                break

        facts['os_version'] = got['kickstart_ver_str']
        facts['chassis_id'] = got['chassis_id']
        facts['virtual'] = bool('NX-OSv' in facts['chassis_id'])

        row = got['TABLE_slot']['ROW_slot']['TABLE_slot_info']['ROW_slot_info'][0]
        facts['serial_number'] = row['serial_num']
        facts['hw_model'] = row['model_num']
        facts['hw_part_number'] = row['part_num']
        facts['hw_part_version'] = row['part_revision']
        facts['hw_version'] = row['hw_ver']

        got = exec_show('show interface mgmt0')
        raw_mac = got['TABLE_interface']['ROW_interface']['eth_hw_addr'].replace('.', '')
        facts['mac_address'] = ':'.join(s.encode('hex') for s in raw_mac.decode('hex'))

    def __getattr__(self, item):
        # ##
        # ## this is rather perhaps being a bit "too clever", but I sometimes
        # ## can't help myself ;-P
        # ##

        # if the named addon module does not exist, this will raise
        # an ImportError.  We might want to catch this and handle differently

        mod = importlib.import_module('.autoload.%s' % item, package=__package__)

        def wrapper(*vargs, **kvargs):
            cls = getattr(mod, '_%s' % item)
            self.__dict__[item] = cls(self, *vargs, **kvargs)

        return wrapper



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

from collections import namedtuple
import time
import logging
from functools import partial
from aeon.nxos.exceptions import CommandError

_MODEL_RC_LIMITS = {

}


class _guestshell(object):
    GUESTSHELL_CPU = 6
    GUESTSHELL_DISK = 1024
    GUESTSHELL_MEMORY = 3072

    Resources = namedtuple('Resources', 'cpu memory disk')

    def __init__(self, device,
                 cpu=GUESTSHELL_CPU, memory=GUESTSHELL_MEMORY,
                 disk=GUESTSHELL_DISK, log=None):

        self.device = device
        self.guestshell = partial(self.device.api.exec_opcmd, msg_type='cli_show_ascii')
        self.cli = self.device.api.exec_opcmd
        self.log = log or logging.getLogger()

        self.sz_max = {}
        self._get_sz_max()

        self.sz_need = _guestshell.Resources(
            cpu=min(cpu, self.sz_max['cpu']),
            memory=min(memory, self.sz_max['memory']),
            disk=min(disk, self.sz_max['disk']))

        self.sz_has = None
        self._state = None
        self.exists = False

    # ---------------------------------------------------------------
    # -----
    # -----                     PROPERTIES
    # -----
    # ---------------------------------------------------------------

    @property
    def state(self):
        cmd = 'show virtual-service detail name guestshell+'
        try:
            got = self.cli(cmd)
        except CommandError:
            # means there is no guestshell
            self.exists = False
            self._state = 'None'
            return self._state

        try:
            self._state = got['TABLE_detail']['ROW_detail']['state']
            return self._state
        except TypeError:
            # means there is no guestshell
            self.exists = False
            self._state = 'None'
            return self._state

    @property
    def size(self):
        self._get_sz_info()
        return self.sz_has

    # ---------------------------------------------------------------
    # -----
    # -----                     PUBLIC METHODS
    # -----
    # ---------------------------------------------------------------

    def setup(self):
        self.log.info("/START(guestshell): setup")

        state = self.state
        self.log.info("/INFO(guestshell): current state: %s" % state)

        if 'Activated' == state:
            self._get_sz_info()
            if self.sz_need != self.sz_has:
                self.log.info("/INFO(guestshell): need to resize, please wait...")
                self.resize()
                self.reboot()
        else:
            self.log.info(
                "/INFO(guestshell): not activated, enabling with proper size, "
                "please wait ...")
            self.resize()
            self.enable()

        self._get_sz_info()
        self.log.info("/END(guestshell): setup")

    def reboot(self):
        self.guestshell('guestshell reboot')
        self._wait_state('Activated')

    def enable(self):
        self.guestshell('guestshell enable')
        self._wait_state('Activated')

    def destroy(self):
        if 'None' == self.state:
            return

        if 'Activating' == self.state:
            self._wait_state('Activated')

        if 'Deactivating' == self.state:
            self._wait_state('Deactivated')

        self.guestshell('guestshell destroy')

        self._wait_state('None')

    def disable(self):
        self.guestshell('guestshell disable')
        self._wait_state('Deactivated')

    def resize_cpu(self, cpu):
        value = min(cpu, self.sz_max['cpu'])
        self.guestshell('guestshell resize cpu {}'.format(value))

    def resize_memory(self, memory):
        value = min(memory, self.sz_max['memory'])
        self.guestshell('guestshell resize memory {}'.format(value))

    def resize_disk(self, disk):
        value = min(disk, self.sz_max['disk'])
        self.guestshell('guestshell resize rootfs {}'.format(value))

    def resize(self):
        self.resize_cpu(self.sz_need.cpu)
        self.resize_memory(self.sz_need.memory)
        self.resize_disk(self.sz_need.disk)

    def run(self, command):
        return self.guestshell('guestshell run %s' % command)

    def sudoers(self, enable):
        """
        This method is used to enable/disable bash sudo commands running
        through the guestshell virtual service.  By default sudo access
        is prevented due to the setting in the 'sudoers' file.  Therefore
        the setting must be disabled in the file to enable sudo commands.

        This method assumes that the "bash-shell" feature is enabled.
        @@@ TO-DO: have a mech to check &| control bash-shell feature support

        :param enable:
            True - enables sudo commands
            False - disables sudo commands

        :return:
            returns the response of the sed command needed to make the
            file change
        """
        f_sudoers = "/isan/vdc_1/virtual-instance/guestshell+/rootfs/etc/sudoers"

        if enable is True:
            sed_cmd = r" 's/\(^Defaults *requiretty\)/#\1/g' "
        elif enable is False:
            sed_cmd = r" 's/^#\(Defaults *requiretty\)/\1/g' "
        else:
            raise RuntimeError('enable must be True or False')

        self.guestshell("run bash sudo sed -i" + sed_cmd + f_sudoers)

    # ---------------------------------------------------------------
    # -----
    # -----                     PRIVATE METHODS
    # -----
    # ---------------------------------------------------------------

    def _get_sz_max(self):
        got = self.cli('show virtual-service global')
        limits = got['TABLE_resource_limits']['ROW_resource_limits']
        for resource in limits:
            name = resource['media_name']
            max_val = int(resource['quota'])
            if 'CPU' in name:
                self.sz_max['cpu'] = max_val
            elif 'memory' in name:
                self.sz_max['memory'] = max_val
            elif 'flash' in name:
                self.sz_max['disk'] = max_val

    def _get_sz_info(self):
        """
        Obtains the current resource allocations, assumes that the
        guestshell is in an 'Activated' state
        """
        if 'None' == self._state:
            return None

        cmd = 'show virtual-service detail name guestshell+'
        got = self.cli(cmd)
        got = got['TABLE_detail']['ROW_detail']

        sz_cpu = int(got['cpu_reservation'])
        sz_disk = int(got['disk_reservation'])
        sz_memory = int(got['memory_reservation'])

        self.sz_has = _guestshell.Resources(
            cpu=sz_cpu, memory=sz_memory, disk=sz_disk)

    def _wait_state(self, state, timeout=60, interval=1, retry=0):
        now_state = None
        time.sleep(interval)

        while timeout:
            now_state = self.state
            if now_state == state:
                return
            time.sleep(interval)
            timeout -= 1

        if state == 'Activated' and now_state == 'Activating':
            # maybe give it some more time ...

            if retry > 2:
                msg = '/INFO(guestshell): waiting too long for Activated state'
                self.log.critical(msg)
                raise RuntimeError(msg)

            self.log.info('/INFO(guestshell): still Activating ... giving it some more time')
            self._wait_state(state, retry + 1)

        else:
            msg = '/INFO(guestshell): state %s never happened, still %s' % (state, now_state)
            self.log.critical(msg)
            raise RuntimeError(msg)



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import os
import argparse
import logging

import aeon.exceptions as exceptions


class ArgumentParser(argparse.ArgumentParser):
    class ParserError(Exception):
        pass

    def error(self, message):
        raise ArgumentParser.ParserError(message)


class Stdargs(object):
    _ENV = {
        'TARGET_USER': 'AEON_TUSER',
        'TARGET_PASSWD': 'AEON_TPASSWD',
        'TARGET': 'AEON_TARGET',
        'LOGFILE': 'AEON_LOGFILE'
    }

    def __init__(self, **kwargs):
        self.progname = kwargs.get('name', 'aeon-nxos')
        self.psr = ArgumentParser(**kwargs)

        self.args = None
        self.target = None
        self.user = None
        self.passwd = None
        self.log = None

        self._setup_stdargs()

    def _setup_stdargs(self):
        self.psr.add_argument(
            '-t', '--target',
            default=os.getenv(Stdargs._ENV['TARGET']),
            help='Target hostname or ip-addr')

        self.psr.add_argument(
            '--logfile',
            default=os.getenv(Stdargs._ENV['LOGFILE']),
            help='name of log file')

        self.psr.add_argument(
            '--json',
            action='store_true', default=True,
            help='output in JSON')

        group = self.psr.add_argument_group('authentication')

        group.add_argument(
            '-u', '--user',
            help='Target username')

        group.add_argument(
            '-U', dest='env_user',
            default=Stdargs._ENV['TARGET_USER'],
            help='Target username environment variable')

        group.add_argument(
            '-P', dest='env_passwd',
            default=Stdargs._ENV['TARGET_PASSWD'],
            help='Target password environment variable')

    def parse_args(self):
        self.args = self.psr.parse_args()

        self.target = self.args.target
        self.user = self.args.user or os.getenv(self.args.env_user)
        self.passwd = os.getenv(self.args.env_passwd)

        if not self.target:
            raise exceptions.TargetError('missing target value')
        if not self.user:
            raise exceptions.TargetError('missing username value')
        if not self.passwd:
            raise exceptions.TargetError('missing password value')

        self.log = logging.getLogger(name=self.progname)
        if self.args.logfile:
            self._setup_logging()

        return self.args

    def _setup_logging(self, level=logging.INFO):
        self.log.setLevel(level)
        fh = logging.FileHandler(self.args.logfile)

        fmt = logging.Formatter(
            '%(asctime)s:%(levelname)s: {target}:%(message)s'
            .format(target=self.args.target))

        fh.setFormatter(fmt)
        self.log.addHandler(fh)



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

from aeon.nxos.device import Device  # NOQA



#
#  Copyright 2016-present, Apstra, Inc.  All rights reserved.
#
#  This source code is licensed under Community End User License Agreement
#  found in the LICENSE.txt file in the root directory of this source tree.
#

from setuptools import setup, find_packages
from glob import glob

# parse requirements
req_lines = [line.strip() for line in open(
    'requirements.txt').readlines()]
install_reqs = list(filter(None, req_lines))

libdir = 'pylib'
packages = find_packages(libdir)

setup(
    name="aeon-venos",
    namespace_packages=['aeon'],
    version="0.5.7",
    author="Jeremy Schulman",
    url='https://github.com/Apstra/aeon-venos',
    author_email="jeremy@apstra.com",
    description="Aeon vendor NOS driver library",
    license="Apache 2.0",
    keywords="networking automation vendor-agnostic",
    package_dir={'': libdir},
    packages=packages,
    install_requires=install_reqs,
    scripts=glob('bin/*'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
    ],
)



# Copyright 2014-present, Apstra, Inc. All rights reserved.
#
# This source code is licensed under End User License Agreement found in the
# LICENSE file at http://www.apstra.com/community/eula

import socket
import paramiko
from aeon.exceptions import LoginNotReadyError


__all__ = ['Connector']


class Connector(object):
    DEFAULT_PROTOCOL = 'ssh'

    def __init__(self, hostname, **kwargs):
        self.hostname = hostname
        self.proto = kwargs.get('proto') or self.DEFAULT_PROTOCOL
        self.port = kwargs.get('port') or socket.getservbyname('ssh')
        self.user = kwargs.get('user')
        self.passwd = kwargs.get('passwd')

        self._client = paramiko.SSHClient()
        self._client.load_system_host_keys()
        self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        self.open()

    def open(self):
        try:
            self._client.connect(
                self.hostname, port=self.port,
                username=self.user, password=self.passwd)

        except Exception as exc:
            raise LoginNotReadyError(exc=exc, message='Unable to connect.')

    def close(self):
        self._client.close()

    def execute(self, commands, stop_on_error=True):
        results = []
        exit_code_collector = 0

        for cmd in commands:
            stdin, stdout, stderr = self._client.exec_command(cmd)
            exit_code = stdout.channel.recv_exit_status()
            exit_code_collector |= exit_code

            results.append(dict(cmd=cmd, exit_code=exit_code,
                                stdout=stdout.read(),
                                stderr=stderr.read()))

            if stop_on_error is True and exit_code != 0:
                return False, results

        return bool(0 == exit_code_collector), results
